package com.nokia.pm_interface_5g.task;

import com.jcraft.jsch.JSchException;
import com.jcraft.jsch.SftpException;
import com.nokia.common.gpload.GploadUtil;
import com.nokia.common.gpload.entity.GploadResult;
import com.nokia.common.ssh.SSHUtil;
import com.nokia.common.ssh.exception.SSHUtilException;
import com.xxl.job.core.context.XxlJobHelper;
import com.xxl.job.core.handler.annotation.XxlJob;
import lombok.Data;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.csv.CSVFormat;
import org.apache.commons.csv.CSVParser;
import org.apache.commons.csv.CSVPrinter;
import org.apache.commons.csv.CSVRecord;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;
import org.springframework.util.CollectionUtils;
import org.springframework.util.StringUtils;

import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

@Data
@Component
@Slf4j
public class FiveGPmTask {

    @Value("${fiveg.pm.download.host:10.17.180.55}")
    private String host;
    @Value("${fiveg.pm.download.port:22}")
    private Integer port;
    @Value("${fiveg.pm.download.username:nokia}")
    private String username;
    @Value("${fiveg.pm.download.password:Nokia*123}")
    private String password;
    @Value("${fiveg.pm.download.sourceDir:/data/out2/pm_5g_hour}")
    private String sourceDir;
    @Value("${fiveg.pm.download.targetDir:download/}")
    private String downloadTargetDir;
    @Value("${fiveg.pm.filePrefix:pm_5g_hour_}")
    private String filePrefix;
    @Value("${fiveg.pm.distinct.targetDir:distinct/}")
    private String distinctTargetDir;
    private final DateFormat dateFormat = new SimpleDateFormat("yyyyMMddHH");

    /**
     * 定时任务
     *
     */
//    @Scheduled(cron = "0 */10 * * * ?")
    @XxlJob("pm5gJobHandler")
    public void pm5gJobHandler() {
        try {
            // 创建文件夹
            Files.createDirectories(Paths.get(downloadTargetDir));
            Files.createDirectories(Paths.get(distinctTargetDir));
            SSHUtil sshUtil = new SSHUtil(host, port, username, password);
            // 获取文件列表
            List<String> list = sshUtil.ls(sourceDir);
            if (CollectionUtils.isEmpty(list)) {
                log.debug("没有文件");
                XxlJobHelper.log("没有文件");
                return;
            }

            log.debug("扫描到的文件: {}", list);
            XxlJobHelper.log("扫描到的文件: {}", list);
            for (String t : list) {
                singleTask(t);
            }
        } catch (Exception e) {
            log.error("发生异常了: {}", e.getMessage(), e);
            XxlJobHelper.log("发生异常了: {}", e.getMessage(), e);
            XxlJobHelper.handleFail(e.getMessage());
        }
    }

    /**
     * 单一任务
     *
     * @param filename 文件名
     */
    public void singleTask(String filename) throws JSchException, SSHUtilException, IOException, SftpException {
        download(filename);
        distinct(filename);
        gpload(filename);
    }

    /**
     * 下载文件
     *
     * @param filename 文件名
     */
    public void download(String filename) throws JSchException, SSHUtilException, IOException, SftpException {
        SSHUtil sshUtil = new SSHUtil(host, port, username, password);
        String sourceFilePath = sourceDir + "/" + filename;
        File targetFile = new File(downloadTargetDir + filename);
        if (!targetFile.exists()) {
            targetFile.getParentFile().mkdirs();
        }
        String targetPath = targetFile.getAbsolutePath();
        boolean b = sshUtil.scpFrom(sourceFilePath, targetPath);
        if (b) {
            log.debug("文件 {} 下载成功...", targetPath);
            XxlJobHelper.log("文件 {} 下载成功...", targetPath);
            //删除远程文件
            sshUtil.delete(sourceFilePath);
        }
    }

    /**
     * 去重
     *
     * @param filename 文件名
     */
    public void distinct(String filename) throws IOException {
        log.debug("文件 {} 去重...", filename);
        XxlJobHelper.log("文件 {} 去重...", filename);
        String inputFilePath = downloadTargetDir + filename;
        String outputFilePath = distinctTargetDir + filename;
        Path inputPath = Paths.get(inputFilePath);
        try (CSVParser parser = CSVFormat.DEFAULT.builder().build()
                .parse(new InputStreamReader(Files.newInputStream(inputPath), StandardCharsets.UTF_8));
             OutputStreamWriter osw = new OutputStreamWriter(Files.newOutputStream(Paths.get(outputFilePath)),
                     StandardCharsets.UTF_8);
             CSVPrinter printer = new CSVPrinter(osw, CSVFormat.DEFAULT);) {
            Map<String, CSVRecord> map = new HashMap<>();
            // 去重
            for (CSVRecord t : parser) {
                map.put(t.get(2), t);
            }
            // 重排
            rerrange(printer, map);
        }

        log.debug("文件 {} 去重完成...", filename);
        XxlJobHelper.log("文件 {} 去重完成...", filename);
        // 删除本地源文件
        Files.deleteIfExists(inputPath);
        log.debug("删除本地源文件 {}...", filename);
        XxlJobHelper.log("删除本地源文件 {}...", filename);
    }

    public void gpload(String filename) throws IOException {
        String gploadCommand = "sh /data1/pm_5g/gpload/pm_nr_gpload.sh " + filename;
        GploadResult gpload = GploadUtil.gpload(gploadCommand);
        if (Boolean.TRUE.equals(gpload.getTaskStatus())) {
            log.debug("gpload完成: {}", gpload);
            XxlJobHelper.log("gpload完成: {}", gpload);
            // 删除重排文件
            Files.deleteIfExists(Paths.get(distinctTargetDir + filename));
        } else {
            log.error("gpload 失败: {}", gpload.getMessage());
            XxlJobHelper.log("gpload 失败: {}", gpload.getMessage());
        }
    }

    /**
     * 保留4位小数
     */
    BigDecimal format(String s) {
        if (!StringUtils.hasText(s)) {
            return null;
        }

        return new BigDecimal(s).setScale(4, RoundingMode.UP);
    }

    /**
     * 重排
     */
    private void rerrange(CSVPrinter printer, Map<String, CSVRecord> map) throws IOException {
        for (CSVRecord t : map.values()) {
            String nciStr = t.get(2);
            String enbid = "";
            String cellId = "";
            if (!StringUtils.hasText(nciStr)) {
                continue;
            }

            if (nciStr.contains(".")) {
                String[] a = nciStr.split("\\.");
                enbid = a[1];
                cellId = a[2];
            } else if (nciStr.contains("-")) {
                String[] a = nciStr.split("-");
                enbid = a[1];
                cellId = a[2];
            }

            if (!StringUtils.hasText(enbid) || !StringUtils.hasText(cellId)) {
                continue;
            }
            String enbCell = enbid + "_" + cellId;
            int nci = Integer.parseInt(enbid) * 4096 + Integer.parseInt(cellId);
            printer.printRecord(
                    t.get(1),
                    t.get(3),
                    nci,
                    enbid,
                    cellId,
                    enbCell,
                    t.get(6),
                    format(t.get(10)),
                    format(t.get(18)),
                    format(t.get(14)),
                    format(t.get(23)),
                    format(t.get(79)),
                    format(t.get(59)),
                    format(t.get(60)),
                    format(t.get(53)),
                    format(t.get(56)),
                    format(t.get(80)),
                    format(t.get(11)),
                    format(t.get(81)),
                    format(t.get(82)),
                    format(t.get(83)),
                    format(t.get(84)),
                    format(t.get(85)),
                    format(t.get(86)),
                    format(t.get(87)),
                    format(t.get(88)),
                    format(t.get(89)),
                    format(t.get(90)),
                    format(t.get(91)),
                    format(t.get(92)),
                    format(t.get(93)),
                    format(t.get(94)),
                    format(t.get(95)),
                    format(t.get(96)),
                    format(t.get(97)),
                    format(t.get(98)),
                    format(t.get(99)),
                    format(t.get(100)),
                    format(t.get(101)),
                    format(t.get(102)),
                    format(t.get(103)),
                    format(t.get(104)),
                    format(t.get(62)),
                    format(t.get(63)),
                    format(t.get(105)),
                    format(t.get(106)),
                    format(t.get(57)),
                    format(t.get(58)),
                    format(t.get(107)),
                    format(t.get(108)),
                    format(t.get(109)),
                    format(t.get(110)),
                    t.get(111),
                    t.get(112),
                    t.get(113),
                    t.get(8),
                    t.get(7),
                    format(t.get(20)),
                    format(t.get(19)),
                    format(t.get(12)),
                    format(t.get(13)),
                    format(t.get(114)),
                    format(t.get(115)),
                    format(t.get(116)),
                    t.get(117),
                    t.get(118),
                    t.get(119),
                    t.get(120),
                    t.get(121),
                    t.get(122),
                    t.get(123),
                    t.get(124),
                    format(t.get(125)),
                    format(t.get(126)),
                    format(t.get(59)),
                    format(t.get(60)),
                    format(t.get(127)),
                    format(t.get(128)),
                    format(t.get(129)),
                    format(t.get(130)),
                    format(t.get(131)),
                    format(t.get(132)),
                    format(t.get(133)),
                    t.get(134),
                    t.get(135),
                    t.get(136),
                    t.get(137),
                    t.get(138),
                    t.get(139),
                    t.get(140),
                    t.get(141),
                    format(t.get(142)),
                    format(t.get(143)),
                    t.get(144),
                    t.get(145),
                    t.get(147),
                    t.get(148),
                    t.get(149),
                    t.get(150),
                    t.get(151),
                    t.get(152),
                    t.get(153),
                    t.get(154),
                    t.get(155),
                    t.get(156),
                    t.get(157),
                    t.get(158),
                    t.get(159),
                    t.get(160),
                    format(t.get(161)),
                    format(t.get(162)),
                    format(t.get(163)),
                    format(t.get(164)),
                    format(t.get(165)),
                    format(t.get(57)),
                    format(t.get(58)),
                    format(t.get(41)),
                    format(t.get(166)),
                    t.get(167),
                    t.get(168),
                    format(t.get(169)),
                    format(t.get(65)),
                    format(t.get(64)),
                    format(t.get(61))
            );
        }
    }
}
